//
// Frdric Boulanger
// Suplec - Dpartement Informatique
// 3 rue Joliot-Curie
// 91192 Gif-sur-Yvette cedex, France
//
// frederic.boulanger@supelec.fr
//
/**
 * La notation ABC reprsente les notes par une lettre :
 *   A = la
 *   B = si
 *   C = do
 *   D = r
 *   E = mi
 *   F = fa
 *   G = sol
 * Un bmol est indiqu par un '_' plac avant la note.
 * Un dise est indiqu par un '^' plac avant la note.
 * L'octave dans laquelle se trouve la note est indique
 * par la casse du nom et par les caractres ''' et ',' situs
 * aprs le nom de la note. Les changements d'octave se font 
 * sur la note do.
 * "C" dsigne le do situ une ligne en-dessous de la clef de sol, 
 * et "B" dsigne le si situ sur la 3e ligne de la clef de sol.
 * Le do qui se situe juste au dessus est not "c".
 * Chaque apostrophe ''' place aprs le nom de la note la monte
 * d'une octave. Chaque virgule ',' place aprs le nom de la note
 * la descend d'une octave. Ainsi, "C,," dsigne le do grave du 
 * violoncelle, situ sur la deuxime ligne en-dessous de la
 * clef de fa, alors que "_e'" dsigne le mi bmol aigu situ
 * sur la 3e ligne au-dessus de la clef de sol.
 */
class NotationABC implements NotationMusicale {
  // Valeur d'un demi-ton = 1/12 d'octave
  private static final double demiton_ = Math.pow(2, 1.0/12); // racine 12e de 2
  // Frquence du la4
  private final double diapason_;
  // Frquence du do "du milieu"
  private final double C4_;

  /**
   * Cre une nouvelle notation ABC avec <code>diapason</code>
   * comme frquence du la not "A" (2e interligne de la clef de sol).
   * @param diapason est la frquence en Hertz du la du 2e interligne 
   *        de la clef de sol.
   */
  public NotationABC(double diapason) {
    diapason_ = diapason;
    // On dtermine la frquence du do qui est  
    // une tierce (3 demi-tons) moins une octave du la
    C4_ = diapason_ * Math.pow(demiton_, 3) / 2;
  }

  /**
   * Dtermine la frquence d'une note note en ABC.
   */
  public double frequence(String note) {
    // On commence par dterminer  combien de demi-tons
    // du do C4 se trouve la note.
    int offsetC = demiTonsDo(note);
    double freq = C4_;
    if (offsetC > 0) {
      // Si la note est au-dessus
      while (offsetC >= 12) {
        // on multiplie sa frquence par 2 pour chaque octave
        freq *= 2;
        offsetC -= 12;
      }
      while (offsetC > 0) {
        // puis on la multiplie par 1/12 d'octave pour chaque demi-ton
        freq *= demiton_;
        offsetC --;
      }
    } else if (offsetC < 0) {
      // Si la note est en-dessous du do, on fait l'inverse
      while (offsetC <= -12) {
        freq /= 2;
        offsetC += 12;
      }
      while (offsetC < 0) {
        freq /= demiton_;
        offsetC ++;
      }
    }
    return freq;
  }

  /**
   * Dtermine  combien de demi-tons du do C4 se trouve
   * une note note en ABC. 
   */
  private int demiTonsDo(String note) {
    int offsetC = 0;
    int idx = 0;
    while (idx < note.length()) {
      switch (Character.toUpperCase(note.charAt(idx))) {
        case '_' :     // bemol
          offsetC --;
          break;
        case '^' :     // dize
          offsetC ++;
          break;
        case '=' :     // bcarre
          break;
        case '\'' :    // octave suprieure
          offsetC += 12;
          break;
        case ',' :     // octave infrieure
          offsetC -= 12;
          break;
        case 'C' :
          break;
        case 'D' :
          offsetC += 2;
          break;
        case 'E' :
          offsetC += 4;
          break;
        case 'F' :
          offsetC += 5;
          break;
        case 'G' :
          offsetC += 7;
          break;
        case 'A' :
          offsetC += 9;
          break;
        case 'B' :
          offsetC += 11;
          break;
      }
      // Les minuscules indiquent l'octave suprieure
      if (Character.isLowerCase(note.charAt(idx))) {
        offsetC += 12;
      }
      idx++;
    }
    return offsetC;
  }
}
